// ======================================================================
// analyzer.cpp - a program for automatically build a bilingual dictionary
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102300: Benjamin Han <benhdj@cs.cmu.edu> Analyzer now requires a
//         part-of-speech tagged target language corpus; post info is
//         used to optimize the translation.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Added a new GA operator,
//         GA_OP_CREEP (local optimization) - this is only valid under
//         relaxed mode.
// 110500: Benjamin Han <benhdj@cs.cmu.edu> Bug fix for sentence scores,
//         plus new and much more efficient incremental EvalFunc.
// 110900: Benjamin Han <benhdj@cs.cmu.edu> Added a new argument '-s' for
//         continuing running from the stored solution population.
// 111200: Benjamin Han <benhdj@cs.cmu.edu> Split analyzer.cpp and create
//         readdic.cpp; also SOLFILE now stores the operator fitness values
//         at the end
// 021601: Benjamin Han <benhdj@cs.cmu.edu> Version 1.4 (bug fixes).
// 021701: Benjamin Han <benhdj@cs.cmu.edu> Add '-addBest' command-line
//         option.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "analyzer.hpp"

#include "creep.hpp"
#include "eval.hpp"

int main (int argc, char *argv[])
{
  int i;
  Dict d1,d2;
  
  cout<<"* Analyzer v"<<ANALYZER_VERSION<<" by Benjamin Han"<<endl<<endl;

  // ------------------------------------------------------------
  // Parsing command-line options
  // ------------------------------------------------------------

  Settings settings;

  if ((i=settings.parse(argc,argv))!=ERR_SUCCESS) return i;

  if (settings.setBackward)
    cout<<"* Translation direction is L2 -> L1."<<endl;
  else cout<<"* Translation direction is L1 -> L2."<<endl;

  // ------------------------------------------------------------
  // Reading/generating DICTFILE
  // ------------------------------------------------------------

  if ((i=readDictionary(settings,d1,d2))!=ERR_SUCCESS) return i;

  // ------------------------------------------------------------
  // Run Solver (if not setDisplay)
  // ------------------------------------------------------------

  EvalFunc ef12(d1,d2,settings.popSize);
  EvalFunc ef21(d2,d1,settings.popSize);

  CreepFunc cf12(d1,d2,ef12);
  CreepFunc cf21(d2,d1,ef21);

  string str;
  bool backward;

  // both dictionaries enter read mode
  d1.setReadMode();
  d2.setReadMode();

  if (!settings.setDisplay)
    {
      Solver solver;
      Constraints cons;
      
      if (settings.setBackward) cons.resize(d2.getNumWords());
      else cons.resize(d1.getNumWords());
      
      // glean thru a dictionary for 'cons'
      DictEntryToConstraints d2c(cons);
      
      VFile solFile;
      fstream statFile;
      
      cout<<"* Solver: Initializing";
      if (settings.relaxedMode) cout<<" in relaxed mode "<<settings.relaxedMode;
      if (settings.setSolFile) cout<<" by loading "<<SOLFILE;
      cout<<"... "<<flush;
      
      // construct the constraints 'cons'
      if (settings.setBackward) d2.readEach(d2c);
      else d1.readEach(d2c);

      if (settings.setSolFile)
	{
	  try { solFile.open(SOLFILE,VFILE_READ); }
	  catch (Exception &e)
	    {
	      cerr<<"* ERROR: "<<e<<endl;
	      return ERR_OPEN_INPUT_FILE;
	    }

	  solFile>>str>>backward;

	  if (str!=settings.corpusFileName || backward!=settings.setBackward)
	    {
	      cerr<<"* ERROR: solution file "<<SOLFILE<<" mismatches."<<endl;
	      return ERR_SOLFILE_MISMATCH;
	    }

	  statFile.open(STATFILE,ios::out|ios::app);
	  if (!statFile)
	    {
	      cerr<<"* ERROR: Cannot open file "<<STATFILE<<" for writing."<<endl;
	      return ERR_WRITING_STATFILE;
	    }
	  
	  solver.init(cons,settings.relaxedMode,settings.opHistorySize,
		      settings.setBackward?cf21:cf12,
		      settings.setBackward?ef21:ef12,
		      settings.printInt,statFile,solFile);

	  solFile.close();
	}
      else
	{
	  statFile.open(STATFILE,ios::out);
	  if (!statFile)
	    {
	      cerr<<"* ERROR: Cannot open file "<<STATFILE<<" for writing."<<endl;
	      return ERR_WRITING_STATFILE;
	    }

	  solver.init(cons,settings.relaxedMode,settings.popSize,settings.accOpRate,settings.opHistorySize,
		      settings.setBackward?cf21:cf12,settings.setBackward?ef21:ef12,settings.printInt,statFile,
		      settings.setAddBest);
	}

      cout<<"done."<<endl;
      
      solver.printStat();
      
      cout<<"* Solver: Running until "<<settings.maxEpoch<<" epochs or the best score reaches "<<settings.goalScore<<" (stats output to "<<STATFILE<<")... "<<flush;
      solver.run(settings.maxEpoch,settings.goalScore);
      cout<<"done."<<endl<<endl;
      
      statFile.close();
      solver.printStat();
      
      cout<<"* Solver: Writing the solutions to "<<SOLFILE<<"... "<<flush;
      try { solFile.open(SOLFILE,VFILE_WRITE); }
      catch (Exception &e)
	{
	  cerr<<"* ERROR: "<<e<<endl;
	  return ERR_WRITING_SOLFILE;
	}

      // remember the corresponding input file and the translation
      // direction
      solFile<<settings.corpusFileName<<settings.setBackward;
      solver.writeSolutions(solFile);
      solFile.close();
      cout<<"done."<<endl<<endl;
    }
  else
    if (settings.setBackward)
      {
	DictEntryToBestSol d2s2(d2.getNumWords());

	d2.readEach(d2s2);

	cout<<"**********************************************************************"<<endl;
	cout<<"* L2 -> L1                                                            "<<endl;
	cout<<"**********************************************************************"<<endl<<endl;
	
	cout<<"* Initial score (taking the top choices) = "<<ef21(d2s2.getBestInitSolution())<<endl<<endl;
	d2.print(d1);
      }
    else
      {
	DictEntryToBestSol d2s1(d1.getNumWords());

	d1.readEach(d2s1);
      
	cout<<"**********************************************************************"<<endl;
	cout<<"* L1 -> L2                                                            "<<endl;
	cout<<"**********************************************************************"<<endl<<endl;
	
	cout<<"* Initial score (taking the top choices) = "<<ef12(d2s1.getBestInitSolution())<<endl<<endl;
	d1.print(d2);
    }

  return ERR_SUCCESS;
}
