// ======================================================================
// analyzer.hpp - a program for automatically build a bilingual dictionary
// 
// 110900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 111200: Benjamin Han <benhdj@cs.cmu.edu> Add readDictionary().
// 111700: Benjamin Han <benhdj@cs.cmu.edu> Fixing bug: In
//         DictEntryToBestSol::DictEntryToBestSol() set bestInitSol.id to 0
// 021601: Benjamin Han <benhdj@cs.cmu.edu> Version 1.4 (bug fixes).
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef ANALYZER_HPP
#define ANALYZER_HPP

#include "dict.hpp"
#include "solver.hpp"

// the following 3 files are automatically generated by Analyzer
#define DICTFILE "analysis.dic"
#define STATFILE "analysis.stat"
#define SOLFILE  "analysis.sol"

// error codes
#define ERR_SUCCESS                    0
#define ERR_ILLEGAL_OPT                1
#define ERR_INTERNAL_CMDLINE           2
#define ERR_NO_INPUT_FILE              3
#define ERR_ZERO_CUTOFF                4
#define ERR_ILLEGAL_NUM_ITER           5
#define ERR_ILLEGAL_RELAXED_MODE       6
#define ERR_ZERO_POPSIZE               7
#define ERR_ILLEGAL_OP_RATE            8
#define ERR_ILLEGAL_OP_HPPISTORY_SIZE    9
#define ERR_ILLEGAL_PRTINT             10
#define ERR_OPEN_INPUT_FILE            11
#define ERR_WRITING_DICTFILE           12
#define ERR_WRITING_STATFILE           13
#define ERR_WRITING_SOLFILE            14
#define ERR_POST_FILE_MISMATCH         15
#define ERR_TAGS_PARSING               16
#define ERR_ILLEGAL_CORPUS_FILE_FORMAT 17
#define ERR_OP_NO_CREEP                18
#define ERR_SOLFILE_MISMATCH           19
#define ERR_NO_ADDBEST                 20
#define ERR_NO_ADDBEST_SOLFILE         21

// remove the leading spaces and convert to lowercase
void processString (string &str);

void printHelp ();
void printStat (const Dict &d1, const Dict &d2);

struct Settings
{
  int c1,c2;
  int numIter;
  int relaxedMode;
  Solver::size_type popSize;
  int maxEpoch;
  list<int>::size_type opHistorySize;
  int printInt;

  double goalScore;
  
  string corpusFileName,postFileName;

  PDist accOpRate;

  bool setBackward;
  bool setAddBest;
  bool setSolFile;
  bool setDisplay;

  Settings ()
  {
    setBackward=false;
    setAddBest=false;
    setSolFile=false;
    c1=c2=-1;
    numIter=2;
    relaxedMode=0;
    popSize=100;
    goalScore=-1.0;
    maxEpoch=500;
    setDisplay=false;
    opHistorySize=5;
    printInt=1;

    accOpRate.resize(GA_OP_NUM);
  }

  int parse (int argc, char *argv[]);
};

int readDictionary (Settings &settings, Dict &d1, Dict &d2);

class DictEntryToConstraints: public DictEntryFunc
{
  Constraints &cons;

  int id;
  
public:

  // ASSUMPTION: 'cons' is already of the right size
  DictEntryToConstraints (Constraints &cons):cons(cons),id(0) {}

  void operator () (int wid, const DictEntry &de);
};

// for collection the solution consisting of the top choice of each
// candidate concept set
class DictEntryToBestSol: public DictEntryFunc
{
  // the top choices in each candidate concept set are collected to
  // form an initial best solution - this serves as an indicator of
  // how well the individual assignements agree with the global,
  // solution-wise evaluation
  Solution bestInitSol;

public:

  // ASSUMPTION: 'cons' is already of the right size
  DictEntryToBestSol (Solution::size_type size)
  {
    bestInitSol.resize(size);
    bestInitSol.setID(0);       // dummy solution ID: '-d' switch is in effect

    // force a full evaluation
    for (Solution::size_type i=0;i<size;++i) bestInitSol.changed.insert(i);
  }

  // WARNING: be sure to call this only after Dict::readEach() call with
  // this function obj (only then 'bestInitSol' is fully instantiated)
  const Solution &getBestInitSolution () const { return bestInitSol; }

  void operator () (int wid, const DictEntry &de);
};



#endif
