// ======================================================================
// creep.cpp - creeping function (local optimization) for the GA Solver
// 
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 110500: Benjamin Han <benhdj@cs.cmu.edu> Bug fix: the old score of
//         a sentence depends on the particular solution - now it's obtained
//         from EvalFunc::()
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stack>

#include <math.h>

#include "random.hpp"
#include "creep.hpp"

// branching factor of the search for the locally optimal translation
#define BFACTOR       3
#define MAX_NUM_PATHS 500  // only try these # of different translated sentences
#define MAX_PORTION   .5   // only try this portion of all different translations

void CreepFunc::initDFSNode (int sid, int tid, DFSNode &parent, DFSNode &child,
			     const Sentence &s)
{
  child.wid=tid;
  child.depth=parent.depth+1;
  child.tMap=parent.tMap;
  child.tMap.addTagMap(d2.getEntry(tid).tMap);

  child.tList=parent.tList;
  child.tList.push_back(Trans(sid,tid));

  if (child.depth==1)
    {
      child.lScore=0.0;
      child.k=0;
    }
  else if (parent.wid!=UNASSIGNED && child.wid!=UNASSIGNED && 
	   parent.wid!=child.wid)
    {
      child.lScore=parent.lScore+(1.0/s.minDistance(parent.wid,child.wid));
      child.k=parent.k+1;
    }
  else
    {
      child.lScore=parent.lScore;
      child.k=parent.k;
    }
}

bool CreepFunc::operator () (Solution &sol)
{
  Sentence::size_type depth;
  int i,maxNumPaths,sid,j=getRandInt(d1.getNumSentences());
  double oldScore,lScore,pScore,score,temp;
  TagMap tTMap;
  set<int> c;    // choice words
  set<int>::const_iterator ci;
  Range r;
  Range::size_type k;
  stack<DFSNode> dfsStack;
  DFSNode parent,child;

  const Sentence &sSentence=d1.getSentence(j);
  const Sentence &tSentence=d2.getSentence(j);

  if ((temp=pow(BFACTOR,sSentence.size())*(double)MAX_PORTION)>MAX_NUM_PATHS)
    maxNumPaths=MAX_NUM_PATHS;
  else maxNumPaths=(int)temp;

  depth=sSentence.size();
  tTMap=tSentence.tMap;
  oldScore=ef(sol.getID(),j);

  // our job here is to minimize distance
  // for now only a simple approximate algo. - beam DFS search - is used
  // 1. randomly pick BFACTOR candidate concepts for each word in the source language
  //    according to the concept confidences
  // 2. compute the new post distribution distance of the combinations of these
  //    choice words
  // 3. any improvement then report, otherwise report failure

  sid=sSentence[0].wid;
  const CVec &cVec=d1.getEntry(sid).cVec;
  r.resize(cVec.size());

  for (k=0;k<cVec.size();++k)     // silly conversion from CVec to Range
    {
      r[k].val=cVec[k].wordID;
      r[k].p=cVec[k].c;
    }
  selectMulti(r,BFACTOR,c);

  parent.depth=0;

  for (ci=c.begin();ci!=c.end();++ci)
    {
      initDFSNode(sid,*ci,parent,child,tSentence);
      dfsStack.push(child);
    }
  
  for (i=0;!dfsStack.empty();)
    {
      parent=dfsStack.top();
      dfsStack.pop();

      if (parent.depth<depth)
	{
	  sid=sSentence[parent.depth].wid;
	  const CVec &cVec=d1.getEntry(sid).cVec;
	  r.resize(cVec.size());

	  for (k=0;k<cVec.size();++k)     // silly conversion from CVec to Range
	    {
	      r[k].val=cVec[k].wordID;
	      r[k].p=cVec[k].c;
	    }
	  selectMulti(r,BFACTOR,c);

	  for (ci=c.begin();ci!=c.end();++ci)
	    {
	      initDFSNode(sid,*ci,parent,child,tSentence);
	      dfsStack.push(child);
	    }
	}
      else if (i<maxNumPaths)
	{
	  if (sSentence.size()==1)
	    {
	      if (sol[sSentence[0].wid]!=UNASSIGNED &&
		  tSentence.wp.find(sol[sSentence[0].wid])!=tSentence.wp.end())
		lScore=1.0;
	      else lScore=0.0;
	    }
	  else lScore=parent.lScore/(double)parent.k;

	  pScore=1.0-tTMap.computeDistance(parent.tMap)/M_PI_2;

	  score=(lScore+pScore)/2.0;

	  if (score>oldScore)
	    {
	      TList::const_iterator ti;
	      
	      for (ti=parent.tList.begin();ti!=parent.tList.end();++ti)
		if (sol[ti->sid]!=ti->tid) sol[ti->sid]=ti->tid;
	      
	      return true;
	    }
	  ++i;
	}
      else return false;
    }

  return false;
}
