// ======================================================================
// dict.cpp - a dictionary for machine translation from parallel text
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Added code for TagMap/TagVec 
//         structures: these store the frequency of each part-of-speech 
//         tag for a word.
// 110500: Benjamin Han <benhdj@cs.cmu.edu> Bug fix: Sentence::score is
//         removed - different translations have different scores!
// 111700: Benjamin Han <benhdj@cs.cmu.edu> Remove TagVec stuff (TagMap
//         alone is enough).
// 021601: Benjamin Han <benhdj@cs.cmu.edu> Remove the "w1!=w2" assumption
//         in Sentence::minDistance() and make it return -1 if either w1 or
//         w2 is missing in the sentence.
// 031201: Benjamin Han <benhdj@cs.cmu.edu> Added the missing #include
//         <float.h> for DBL_MIN.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <math.h>
#include <float.h>

#include "misc.hpp"
#include "dict.hpp"

using namespace Toolbox;
using namespace std;

int WordSet::addWord (string &word)
{
  iterator wi=find(word);

  if (wi==end())
  {
    wVec.push_back(word);
    (*this)[word]=id;
    return id++;
  }
  else return wi->second;
}

void WordSet::writeToVFile (VFile &vf) const
{
  vf<<wVec;
}

void WordSet::readFromVFile (VFile &vf)
{
  int i,j;

  wVec.clear();
  clear();

  vf>>wVec;
  for (i=0,j=wVec.size();i<j;++i) insert(make_pair(wVec[i],i));
}

void TagMap::normalize ()
{
  double total=0.0;
  iterator mi;

  #ifndef NDEBUG
  normalized=true;
  #endif

  for (mi=begin();mi!=end();++mi) total+=mi->second;
  for (mi=begin();mi!=end();++mi) mi->second/=total;
}

double TagMap::computeDistance (const TagMap &tMap) const
{
  double norm1_2,norm2_2,product,cos_t,dist;
  const_iterator mi1,mi2;

  // if the word has no POS tag (one possible scenario is that a source language
  // word hasn't been assigned to any target language word so the POS dist. is
  // unknown) assign pi/2 (90 degrees) distance to the two TagMaps - note this
  // implicitly give the worst score to an unassigned word; for an assigned
  // word even when the POS distribution is totally different from the truth,
  // the worst distance is still pi/2
  if (size()==0 || tMap.size()==0) return M_PI_2;

  for (norm1_2=product=0.0,mi1=begin();mi1!=end();++mi1)
    {
      norm1_2+=(mi1->second*mi1->second);
      if ((mi2=tMap.find(mi1->first))!=tMap.end())
	product+=(mi1->second*mi2->second);
    }
  
  for (norm2_2=0.0,mi2=tMap.begin();mi2!=tMap.end();++mi2)
    norm2_2+=(mi2->second*mi2->second);

  // sometime the floating point is not that accurate - if cos_t is a little
  // higher than 1.0 then acos() will bumb - so we have the following safety
  // net
  if ((cos_t=product/sqrt(norm1_2*norm2_2))>=1.0) dist=0.0;
  else if (cos_t==0.0) dist=M_PI_2;
  else dist=acos(cos_t);

  assert(dist>=0.0 && dist<=M_PI_2);

  return dist;
}

void TagMap::addTagMap (const TagMap &tMap)
{
  const_iterator ti;
  iterator ti2;

  for (ti=tMap.begin();ti!=tMap.end();++ti)
    if ((ti2=find(ti->first))!=end()) ti2->second+=ti->second;
    else insert(make_pair(ti->first,ti->second));
}

int Sentence::minDistance (int w1, int w2) const
{
  WordPositions::const_iterator wi;
  
  assert(w1!=w2);

  if ((wi=wp.find(w1))==wp.end()) return -1;  // w1 is missing
  const Positions &p1=wi->second;

  if ((wi=wp.find(w2))==wp.end()) return -1;  // w2 is missing
  const Positions &p2=wi->second;

  Positions::const_iterator pi1,pi2;
  int dist,minDist=INT_MAX;

  for (pi1=p1.begin();pi1!=p1.end();++pi1)
    for (pi2=p2.begin();pi2!=p2.end();++pi2)
      if ((dist=abs(*pi1-*pi2))<minDist) minDist=dist;

  return minDist;
}

// set fastTable and sort each cVec according to the wordID order
void Dict::setReadMode ()
{
  #ifndef NDEBUG
  readMode=true;
  #endif

  size_type s=size();
  iterator di;

  // clear the map part of 'ws' and 'ts' to save memory
  ws.clearMap();
  ts.clearMap();

  fastTable.resize(s);
  for (di=begin();di!=end();++di)
    {
      fastTable[di->first]=di;
      sort(di->second.cVec.begin(),di->second.cVec.end(),conceptWordOrder);
    }
}

void Dict::readEach (DictEntryFunc &f) const
{
  assert(readMode==true);
  for (const_iterator di=begin();di!=end();++di) f(di->first, di->second);
}

void Dict::addSentence (int id, string &sentence)
{
  string word;
  string::size_type i,j;
  iterator di;
  TaggedWord tw;
  Sentence s;
  DictEntry de;            // for a new word
  int numWords;
  Positions p;
  WordPositions::iterator wi;

  assert(readMode==false);

  #ifndef NDEBUG
  de.tMap.normalized=false;
  #endif

  // no tags are given: we can only put POST_UNDECIDED into individual
  // TaggedWord
  tw.tid=POST_UNDECIDED;
  de.tMap.insert(make_pair(POST_UNDECIDED,1));
  de.num=1;

  if (sentence[sentence.length()-1]!=' ') sentence+=' ';

  for (i=numWords=0;i<sentence.length();
       i=sentence.find_first_not_of(" \t",j+1))
  {
    word=sentence.substr(i,(j=sentence.find_first_of(" \t",i))-i);
    if (ispunct(word[0]) && word.length()==1) continue;  // ignore punctuations
    tw.wid=ws.addWord(word);

    // idSet is always sorted and only contains unique members because:
    // ASSUMPTION: id is monotonically increasing!
    if ((di=find(tw.wid))==end())
      {
	de.idSet.push_back(id);
	insert(make_pair(tw.wid,de));
	de.idSet.clear();

	// record positions
	p.push_back(numWords);
	s.wp.insert(make_pair(tw.wid,p));
	p.clear();
      }
    else
      {
	// we don't update tMap because this word is not tagged and the
	// only possible tag is POST_UNDECIDED
	di->second.idSet.push_back(id);
	++di->second.num;

	// record positions
	if ((wi=s.wp.find(tw.wid))!=s.wp.end()) wi->second.push_back(numWords);
	else
	  {
	    p.push_back(numWords);
	    s.wp.insert(make_pair(tw.wid,p));
	    p.clear();
	  }
      }

    s.push_back(tw);
    ++numWords;
  }

  s.tMap.insert(make_pair(POST_UNDECIDED,numWords));
  text.push_back(s);
}

void Dict::addSentence (int id, string &sentence, string &tags)
{
  string word,tag;
  string::size_type i,j,k,l,m,n;
  iterator di;
  TaggedWord tw;
  Sentence s;
  DictEntry de;            // for a new word
  TagMap::iterator ti;
  int wPos;
  Positions p;
  WordPositions::iterator wi;

  assert(readMode==false);

  #ifndef NDEBUG
  de.tMap.normalized=false;
  #endif

  de.num=1;

  if (sentence[sentence.length()-1]!=' ') sentence+=' ';

  for (i=k=wPos=0;i<sentence.length() && k<tags.length();
       i=sentence.find_first_not_of(" \t",j+1),
	 k=tags.find_first_not_of(" \t",l+1))
  {
    word=sentence.substr(i,(j=sentence.find_first_of(" \t",i))-i);
    tag=tags.substr(k,(l=tags.find_first_of(" \t",k))-k);
    if (ispunct(word[0]) && word.length()==1) continue;  // ignore punctuations

    m=tag.find_last_of('/');
    if (m==string::npos)
      {
	string idString;
	intToString(id,idString);
	throw (TagsParsingError(idString));
      }
    n=tag.find_first_of(" \t",m+1);
    if (n==string::npos) n=tag.length();
    tag=tag.substr(m+1,n-m-1);

    tw.wid=ws.addWord(word);
    tw.tid=ts.addWord(tag);

    // idSet is always sorted and only contains unique members because:
    // ASSUMPTION: id is monotonically increasing!
    if ((di=find(tw.wid))==end())
      {
	de.idSet.push_back(id);
	de.tMap.insert(make_pair(tw.tid,1));
	insert(make_pair(tw.wid,de));
	de.idSet.clear();
	de.tMap.clear();

	// record positions
	p.push_back(wPos);
	s.wp.insert(make_pair(tw.wid,p));
	p.clear();
      }
    else
      {
	DictEntry &de2=di->second;

	de2.idSet.push_back(id);
	++de2.num;
	if ((ti=de2.tMap.find(tw.tid))==de2.tMap.end())
	  de2.tMap.insert(make_pair(tw.tid,1));
	else ++ti->second;

	// record positions
	if ((wi=s.wp.find(tw.wid))!=s.wp.end()) wi->second.push_back(wPos);
	else
	  {
	    p.push_back(wPos);
	    s.wp.insert(make_pair(tw.wid,p));
	    p.clear();
	  }	
      }

    s.push_back(tw);
    ++wPos;

    // update the sentence's TagMap
    if ((ti=s.tMap.find(tw.tid))==s.tMap.end())
      s.tMap.insert(make_pair(tw.tid,1));
    else ++ti->second;
  }

  text.push_back(s);  
}

void Dict::computeConcepts (Dict &targetD, int cutoff1, int cutoff2, int numPass)
{
  iterator si;         // source
  iterator ti;         // target
  IDSet result;
  IDSet::size_type s1,s2;
  back_insert_iterator<IDSet> ri(result);
  CVec::iterator ci,ci2;
  int pass;
  double conf,sum,temp;
  Concept cpt;

  assert(readMode==false);

  for (si=begin();si!=end();++si)
  {
    DictEntry &sEntry=si->second;
    IDSet &sSet=sEntry.idSet;
    s2=sSet.size();

    sEntry.tMap.normalize();

    for (ti=targetD.begin();ti!=targetD.end();++ti)
    {
      DictEntry &tEntry=ti->second;
      IDSet &tSet=tEntry.idSet;
      
      set_intersection(sSet.begin(),sSet.end(),tSet.begin(),tSet.end(),ri);
      if ((s1=result.size())!=0)
      {
	conf=((double)s1/(double)s2)*((double)s1/(double)tSet.size());

        sEntry.cVec.push_back(Concept(ti->first,conf));
	tEntry.cVec.push_back(Concept(si->first,conf));
        result.clear();   // reset result
      }
    }

    if (numPass==1)
      {
	// sorting according to the confidence ordering (descending)
	sort(sEntry.cVec.begin(),sEntry.cVec.end(),conceptConfDescendingOrder);

	if (cutoff1!=-1)
	  if (cutoff1<(int)sEntry.cVec.size())
	    sEntry.cVec.erase(sEntry.cVec.begin()+cutoff1,sEntry.cVec.end());
      }

    // normalization
    for (ci=sEntry.cVec.begin(),sum=0.0;ci!=sEntry.cVec.end();++ci) sum+=ci->c;
    for (ci=sEntry.cVec.begin();ci!=sEntry.cVec.end();++ci) ci->c/=sum;
  }

  // sort and normalize target confidences
  for (ti=targetD.begin();ti!=targetD.end();++ti)
    {
      CVec &tCVec=ti->second.cVec;

      ti->second.tMap.normalize();

      if (numPass==1)
	{
	  // sorting according to the confidence ordering (descending)
	  sort(tCVec.begin(),tCVec.end(),conceptConfDescendingOrder);
      
	  if (cutoff2!=-1)
	    if (cutoff2<(int)tCVec.size())
	      tCVec.erase(tCVec.begin()+cutoff2,tCVec.end());
	}
      else
	// maintain word-order sort
	sort(tCVec.begin(),tCVec.end(),conceptWordOrder);

      // normalization
      for (ci=tCVec.begin(),sum=0.0;ci!=tCVec.end();++ci) sum+=ci->c;
      for (ci=tCVec.begin();ci!=tCVec.end();++ci) ci->c/=sum;
    }

  // re-compute the confidences numPass-1 times; the last time reduce the
  // size of each candidate concept set to 'cutoff' (or <)

  for (pass=1;pass<numPass;++pass)
    {
      for (si=begin();si!=end();++si)
	{
	  CVec &sCVec=si->second.cVec;
	  
	  cpt.wordID=si->first;
	  for (ci=sCVec.begin();ci!=sCVec.end();++ci)
	    {
	      CVec &dCVec=targetD.find(ci->wordID)->second.cVec;
	      temp=ci->c;
	      ci2=std::lower_bound(dCVec.begin(),dCVec.end(),cpt,conceptWordOrder);
	      ci->c*=ci2->c;
	      ci2->c*=temp;
	    }

	  if (pass==numPass-1)
	    {
	      // last pass
	      // sorting according to the confidence ordering (descending)
	      sort(sCVec.begin(),sCVec.end(),conceptConfDescendingOrder);
	      
	      if (cutoff1!=-1)
		if (cutoff1<(int)sCVec.size())
		  sCVec.erase(sCVec.begin()+cutoff1,sCVec.end());
	    }

	  // normalization
	  for (ci=sCVec.begin(),sum=0.0;ci!=sCVec.end();++ci) sum+=ci->c;
	  for (ci=sCVec.begin();ci!=sCVec.end();++ci) ci->c/=sum;
	}

      for (ti=targetD.begin();ti!=targetD.end();++ti)
	{
	  CVec &tCVec=ti->second.cVec;
	  
	  if (pass==numPass-1)
	    {
	      // last pass
	      // sorting according to the confidence ordering (descending)
	      sort(tCVec.begin(),tCVec.end(),conceptConfDescendingOrder);
	      
	      if (cutoff2!=-1)
		if (cutoff2<(int)tCVec.size())
		  tCVec.erase(tCVec.begin()+cutoff2,tCVec.end());
	    }
	  
	  // normalization
	  for (ci=tCVec.begin(),sum=0.0;ci!=tCVec.end();++ci) sum+=ci->c;
	  for (ci=tCVec.begin();ci!=tCVec.end();++ci) ci->c/=sum;
	}
    }
}

void Dict::writeToVFile (VFile &vf) const
{
  assert(readMode==false);

  vf<<ws<<ts<<text;
  vf<<*((map<int, DictEntry>*)this);
}

void Dict::readFromVFile (VFile &vf)
{
  vf>>ws>>ts>>text;
  vf>>*((map<int, DictEntry>*)this);
}

void Dict::print (const Dict &conceptD) const
{
  Dict::const_iterator di;
  CVec::const_iterator ci;
  TagMap::const_iterator ti;
  CVec cVec;
  double maxTagConf;
  int maxTag=0;  // suppress warning of "use before initialized"

  cout<<"* Display format:"<<endl;
  cout<<"  word: [sentence-wise word freq.; # of candidate concepts] ->"<<endl
      <<"     concept: confidence [POST:confidence]"<<endl
      <<"     ..."<<endl<<endl;

  for (di=begin();di!=end();++di)
    {
      const DictEntry &entry=di->second;

      cout<<"* "<<ws.getWord(di->first)<<" ["
	  <<(double)entry.idSet.size()/(double)text.size()
	  <<"; "<<entry.cVec.size()<<"] ->"<<endl;

      // regardless whether entry.cVec is sorted in word or confidence
      // order, we do the sorting in descending order here again to 
      // avoid confusion
      cVec=entry.cVec;
      sort(cVec.begin(),cVec.end(),conceptConfDescendingOrder);
      
      for (ci=cVec.begin();ci!=cVec.end();++ci)
	{
	  const TagMap &tMap=conceptD.find(ci->wordID)->second.tMap;

	  cout<<"     "<<conceptD.ws.getWord(ci->wordID)<<": "<<ci->c<<" [";

	  for (ti=tMap.begin(),maxTagConf=DBL_MIN;ti!=tMap.end();++ti)
	    if (ti->second>maxTagConf)
	      {
		maxTagConf=ti->second;
		maxTag=ti->first;
	      }

	  cout<<conceptD.ts.getWord(maxTag)<<':'<<maxTagConf<<']'<<endl;
	}
      cout<<endl;  
    }
}
