// ======================================================================
// dict.hpp - a dictionary for machine translation from parallel text
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Added code for TagMap/TagVec 
//         structures: these store the frequency of each part-of-speech 
//         tag for a word; also cleaned up some ptototypes.
// 110500: Benjamin Han <benhdj@cs.cmu.edu> Bug fix: Sentence::score is
//         removed - different translations have different scores!
// 111700: Benjamin Han <benhdj@cs.cmu.edu> Remove TagVec stuff (TagMap
//         alone is enough).
// 021601: Benjamin Han <benhdj@cs.cmu.edu> Remove the "w1!=w2" assumption
//         in Sentence::minDistance() and make it return -1 if either w1 or
//         w2 is missing in the sentence; vfile.h -> vFile.hpp.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef DICT_HPP
#define DICT_HPP

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <list>
#include <map>
#include <algorithm>

#include <assert.h>

#include "vFile.hpp"
#include "except.hpp"

#define POST_UNDECIDED 0        // the part-of-speech tag for a non-tagged word

using namespace Toolbox;
using namespace std;

DEFINE_SIMPLE_EXCEPTION(TagsParsingError,"Tags parsing error at sentence ID %s.");

// ----------------------------------------------------------------------
// Structure definition for WordSet
// ----------------------------------------------------------------------

class WordSet: protected map<string, int>
{
  int id;
  vector<string> wVec;

  public:

  WordSet ():id(0) {}

  const string &getWord (int i) const 
  { 
    assert(i<wVec.size() && i>=0);
    return wVec[i]; 
  }

  // only adds a word if it's not in WordSet;
  // returns the integer 'handle' of the word
  int addWord (string &word);

  void clearMap () { clear(); }

  void readFromVFile (VFile &vf);
  void writeToVFile (VFile &vf) const;
};

// ----------------------------------------------------------------------
// Structure definitions for Concept, IDSet, CVec, and DictEntry
// ----------------------------------------------------------------------

// concept for a word
struct Concept
{
  int wordID;       // word in the other language serving as the concept
  double c;         // confidence

  Concept () {}
  Concept (int wid, double c):wordID(wid),c(c) {}

  void readFromVFile (VFile &vf) { vf>>wordID>>c; }
  void writeToVFile (VFile &vf) const { vf<<wordID<<c; }
};

// for sort(,,op) - ascending wordID order
inline bool conceptWordOrder (const Concept &c1, const Concept &c2)
{
  return c1.wordID<c2.wordID;
}

// for sort(,,op) - descending confidence order
inline bool conceptConfDescendingOrder (const Concept &c1, const Concept &c2)
{
  return c1.c>c2.c;
}

// for sort(,,op) - ascending confidence order
inline bool conceptConfAscendingOrder (const Concept &c1, const Concept &c2)
{
  return c1.c<c2.c;
}

// IDSet is a set of sentence IDs
// ASSUMPTION: it's sorted and it can only contain UNIQUE members!
typedef vector<int> IDSet;
typedef vector<Concept> CVec;  // candidate vector

struct POST
{
  int tid;                     // part-of-speech tag ID
  double c;                    // confidence of this tag

  void readFromVFile (VFile &vf) { vf>>tid>>c; }
  void writeToVFile (VFile &vf) const { vf<<tid<<c; }
};

// for sort(,,op) - ascending tid order
inline bool postWordOrder (const POST &p1, const POST &p2)
{
  return p1.tid<p2.tid;
}

// for sort(,,op) - descending confidence order
inline bool postConfDescendingOrder (const POST &p1, const POST &p2)
{
  return p1.c>p2.c;
}

// recording the occurrence statistics for a tag
struct TagMap: public map<int, double>
{
  #ifndef NDEBUG
  bool normalized;
  TagMap ():normalized(true) {}
  #endif

  void normalize ();

  // compute the distance between two part-of-speech tag distributions
  // WARNING: 'tMap' will be altered!
  double computeDistance (const TagMap &tMap) const;

  // add (overlay) another TagMap statistics to this TagMap
  void addTagMap (const TagMap &tMap);

  void readFromVFile (VFile &vf) 
  { 
    #ifndef NDEBUG
    normalized=true;
    #endif
    
    vf>>*((map<int,double>*)this);
  }

  void writeToVFile (VFile &vf) const 
  { 
    assert(normalized==true);

    vf<<*((map<int,double>*)this);
  }
};

struct DictEntry
{
  IDSet idSet;      // set of sentence IDs where the word occurs
  CVec cVec;

  TagMap tMap;      // a map for POS tags

  int num;          // # of occurrences of the word

  void readFromVFile (VFile &vf) { vf>>idSet>>cVec>>tMap; }
  void writeToVFile (VFile &vf) const { vf<<idSet<<cVec<<tMap; }
};

// a non-modifying function object for Dict::readEach()
class DictEntryFunc
{
public:

  virtual ~DictEntryFunc () {}

  virtual void operator () (int wid, const DictEntry &de)=0;
};

// ----------------------------------------------------------------------
// Structure definitions for Positions, WordPositions, TaggedWord,
// Sentence and Text
// ----------------------------------------------------------------------

typedef list<int> Positions;
typedef map<int,Positions> WordPositions;

struct TaggedWord
{
  int wid;
  int tid;                         // part-of-speech tag ID

  void writeToVFile (VFile &vf) const { vf<<wid<<tid; }
  void readFromVFile (VFile &vf) { vf>>wid>>tid; }
};

struct Sentence: public vector<TaggedWord>
{
  TagMap tMap;                     // count of each tag in the sentence
  WordPositions wp;                // record the positions of each word

  // the min. word distance between two words (they don't have to be in
  // the sentence - in that case minDistance() returns -1)
  int minDistance (int w1, int w2) const;

  // tMap is not saved
  void writeToVFile (VFile &vf) const { vf<<wp<<*((vector<TaggedWord>*)this); }

  // readFromVFile reconstructs tMap
  void readFromVFile (VFile &vf)
  { 
    const_iterator si;
    TagMap::iterator ti;

    vf>>wp>>*((vector<TaggedWord>*)this);

    for (si=begin();si!=end();++si)
      if ((ti=tMap.find(si->tid))==tMap.end()) tMap.insert(make_pair(si->tid,1));
      else ++ti->second;
  }
};

typedef vector<Sentence> Text;

// ----------------------------------------------------------------------
// Structure definition for Dict
// ----------------------------------------------------------------------

// a dictionary of words (word IDs), their occurances and concept coverages
class Dict: protected map<int, DictEntry>
{
  #ifndef NDEBUG
  bool readMode;
  #endif

  WordSet ws;
  WordSet ts;                         // part-of-speech tag set
  vector<const_iterator> fastTable;   // word ID-to-iterator table

  Text text;                          // set of all sentences

  public:

  Dict ()
  {
    #ifndef NDEBUG
    readMode=false;
    #endif

    string undecidedPOST="_UNDECIDED_";

    // ASSUMPTION: "_UNDECIDED_" must be given an numeric ID POST_UNDECIDED,
    // for the initial value of WordSet::id is 0
    ts.addWord(undecidedPOST);
  }

  // ============================================================
  // Read mode functions
  // 1. Must call setReadMode() before using any of these
  // 2. Once entering read mode there's no way back
  // 3. setReadMode() reorders all cVecs according to wordID order
  // ------------------------------------------------------------
  void setReadMode ();
  const DictEntry &getEntry (int i) const
  {
    assert(readMode==true);
    return fastTable[i]->second;
  }

  void readEach (DictEntryFunc &f) const;

  // ============================================================

  const Sentence &getSentence (int sid) const { return text[sid]; }
  const string &getWord (int wid) const { return ws.getWord(wid); }
  const string &getTag (int tid) const { return ts.getWord(tid); }

  size_type getNumWords () const { return size(); }
  int getNumSentences () const { return text.size(); }

  // ============================================================
  // The following functions are NOT compatible with read mode
  // ------------------------------------------------------------
  // 1. add all words of a sentence into the dictionary;
  // 2. id must be monotonically increasing!
  // 3. ignore punctuations!
  // 4. the 2nd version of addSentence() adds the part-of-speech tags 'tags';
  //    in case of parsing errors it throws TagsParsingError.
  void addSentence (int id, string &line);
  void addSentence (int id, string &line, string &tags);

  // 1. Compute candidiate concepts against target dictionary; also normalize
  //    the TagMap of each DictEntry
  // 2. Only retain the top 'cutoff1' choices for the source dict and 
  //    the top 'cutoff2' choices for 'targetD'.
  // 3. 'targetD' is also updated so one call processes both.
  // 4. 'numIter' is the # of iterations to compute the confidences.
  void computeConcepts (Dict &targetD, int cutoff1, int cutoff2, int numIter);

  void writeToVFile (VFile &vf) const;
  // ============================================================

  void readFromVFile (VFile &vf);
  void print (const Dict &conceptD) const;
};


#endif
