// ======================================================================
// eval.cpp - global evaluation of a solution for GA Solver
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102300: Benjamin Han <benhdj@cs.cmu.edu> Changed to base the evaluation
//         on part-of-speech agreement (originally it's based on locality
//         hypothesis).
// 110500: Benjamin Han <benhdj@cs.cmu.edu> Now only evaluates the sentences
//         for which the relevant translations are altered in the solution
// 111700: Benjamin Han <benhdj@cs.cmu.edu> EvalFunc::EvalFunc() doesn't
//         initialize the score of an individual sentence for a translation
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <math.h>

#include "eval.hpp"

#define K 1  // for locality hypothesis

// 'popSize' is the size of the solution population
EvalFunc::EvalFunc (Dict &d1, const Dict &d2, Solver::size_type popSize):d1(d1),d2(d2)
{
  ScoreTable::size_type i;
  int numSen=d1.getNumSentences();
  
  // since in Solver::init() every solution is forced to be evaluated against
  // *all* sentences, individual score of a sentence in sTable[i] needs not be
  // initialized to any value
  sTable.resize(popSize);
  for (i=0;i<popSize;++i) sTable[i].resize(numSen);
}

double EvalFunc::operator () (const Solution &sol)
{
  int i,d,s=d1.getNumSentences();  // d1 is for source language
  double score,lScore,pScore,localScore;
  Sentence::const_iterator si;
  Sentence::size_type j,kappa,size;
  TagMap sTMap,tTMap;
  int w1,w2;

  Scores &scores=sTable[sol.getID()];
  
  for (i=0,score=0.0;i<s;++i,sTMap.clear(),tTMap.clear())
    {
      const Sentence &sSentence=d1.getSentence(i);
      const Sentence &tSentence=d2.getSentence(i);

      for (si=sSentence.begin();
	   si!=sSentence.end() && sol.isNotChanged(si->wid);++si);

      // no relevant translation is changed - skip this sentence
      if (si==sSentence.end()) 
	{
	  score+=scores[i];
	  continue;
	}

      // ---------- compute localilty score ----------
      
      lScore=0.0;
      if ((kappa=abs((int)tSentence.size()-(int)K))<K) kappa=K;
      if ((size=sSentence.size())==1)
	{
	  if (sol[sSentence[0].wid]!=UNASSIGNED &&
	      tSentence.wp.find(sol[sSentence[0].wid])!=tSentence.wp.end())
	    lScore=1.0;
	}
      else
	{
	  for (j=0;j<size-K;++j)
	    {
	      w1=sol[sSentence[j].wid];
	      w2=sol[sSentence[j+K].wid];
	      
	      if (w1!=UNASSIGNED && w2!=UNASSIGNED &&
		  (d=tSentence.minDistance(w1,w2))!=-1)
		lScore+=(kappa-abs(d-K))/kappa;
	    }
	  
	  // average lScore
	  lScore/=(double)(size-K);

	  assert(lScore<=1.0 && lScore>=0.0);
	}

      // ---------- compute POST score ----------

      tTMap=tSentence.tMap;

      // sTMap contains the total confidence of each part-of-speech
      // tag for the whole sentence in the *source* language
      for (si=sSentence.begin();si!=sSentence.end();++si)
	if (sol[si->wid]!=UNASSIGNED)
	  sTMap.addTagMap(d2.getEntry(sol[si->wid]).tMap);

      // reverse the optimization direction
      pScore=1.0-sTMap.computeDistance(tTMap)/M_PI_2;

      assert(pScore<=1.0 && pScore>=0.0);

      localScore=(lScore+pScore)/2.0;  // for now it's equal weight
      scores[i]=localScore;

      score+=localScore;
    }

  return score;
}
