// ======================================================================
// printsol.cpp - print out the solutions file generated by analyzer
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Now print the Part-of-Speech
//         tag of the translated word also.
// 111700: Benjamin Han <benhdj@cs.cmu.edu> Remove TagVec stuff (TagMap
//         alone is enough).
// 021601: Benjamin Han <benhdj@cs.cmu.edu> vfile.h -> vFile.hpp.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdlib.h>
#include <getopt.h>
#include <float.h>

#include <iostream>
#include <algorithm>

#include "vFile.hpp"
#include "dict.hpp"
#include "solver.hpp"

// default dictionary file
#define DICTFILE "analysis.dic"

// error codes
#define ERR_SUCCESS             0
#define ERR_ILLEGAL_OPT         1
#define ERR_INTERNAL_CMDLINE    2
#define ERR_NO_INPUT_FILE       3
#define ERR_ILLEGAL_START       4
#define ERR_ILLEGAL_END         5
#define ERR_ILLEGAL_RANGE       6
#define ERR_OPEN_INPUT_FILE     7

DEFINE_SIMPLE_EXCEPTION(SolFileMismatch,"The dict file doesn\'t match with the solution file.");

void printHelp ()
{
  cerr<<"* printsol [-start|-s <int>] [-end|-e <int>] [-multi|-m] [-dict|-d <dict file>] sol-file"<<endl<<endl;

  cerr<<"  (default values are shown in square brackets)"<<endl;
  cerr<<"  -start:   starting index [1]"<<endl;
  cerr<<"  -end:     ending index [start]"<<endl;
  cerr<<"  -multi:   showing multiple language 1 words vs. one language 2 word"<<endl;
  cerr<<"  -dict:    dictionary file ["<<DICTFILE<<"]"<<endl;
  cerr<<"  sol-file: solutions file generated from Analyzer"<<endl<<endl;
}

void printStat (const Dict &d1, const Dict &d2)
{
  cout<<"* Total # of sentences = "<<d1.getNumSentences()<<endl;
  cout<<"* Total # of the 1st language words = "<<d1.getNumWords()<<endl;
  cout<<"* Total # of the 2nd language words = "<<d2.getNumWords()<<endl<<endl;
}

void printSol (int start, int end, bool multi, const Dict &d1, const Dict &d2, 
	       VFile &solFile)
{
  int i,j,k;
  bool setBackward;
  vector<Solution> population;
  Concept cpt;
  CVec dCVec;
  TagMap::const_iterator ti;
  int maxTag=0;  // suppress warning of "use before initialized"
  double maxTagConf;

  solFile>>setBackward>>population;
  
  k=population[0].size();   // solution size

  for (i=start-1;i<end;++i)
    {
      cout<<"**********************************************************************"<<endl;
      cout<<" Solution "<<i<<" [numSVars = "<<population[i].getNumSVars()
	  <<"; score = "<<population[i].getScore()<<"]"<<endl;
      cout<<"**********************************************************************"<<endl<<endl;
      
      if (setBackward)
	cout<<"word in L2\tword in L1";
      else cout<<"word in L1\tword in L2";

      cout<<"\tPOST\tconfidence\trank"<<endl;

      for (j=0;j<k;++j)
	{
	  cout<<(setBackward?d2.getWord(j):d1.getWord(j))<<'\t';
	  if (population[i][j]!=UNASSIGNED) 
	    {
	      const TagMap &tMap=(setBackward?
				  d1.getEntry(population[i][j]).tMap:
				  d2.getEntry(population[i][j]).tMap);

	      // I know it's slow but heck it's for printing anyway
	      dCVec=(setBackward?d2.getEntry(j).cVec:d1.getEntry(j).cVec);

	      cout<<(setBackward?d1.getWord(population[i][j]):d2.getWord(population[i][j]))<<'\t';
	      for (ti=tMap.begin(),maxTagConf=DBL_MIN;ti!=tMap.end();++ti)
		if (ti->second>maxTagConf)
		  {
		    maxTagConf=ti->second;
		    maxTag=ti->first;
		  }
	      cout<<(setBackward?
		     d1.getTag(maxTag):d2.getTag(maxTag))<<'\t';

	      cpt.wordID=population[i][j];
	      cout<<(cpt.c=lower_bound(dCVec.begin(),dCVec.end(),cpt,conceptWordOrder)->c)<<'\t';

	      sort(dCVec.begin(),dCVec.end(),conceptConfAscendingOrder);

	      cout<<dCVec.end()-lower_bound(dCVec.begin(),dCVec.end(),cpt,conceptConfAscendingOrder)<<endl;
	    }
	  else cout<<"--UNASSIGNED--"<<endl;
	}

      cout<<endl;
    }
}

int main (int argc, char *argv[])
{
  int opt,numIter,c1,c2;
  int longOptID = 0;
  static struct option longOpts[] =
    {
      {"start", 1, 0, 's'},
      {"end", 1, 0, 'e'},
      {"multi", 1, 0, 'm'},
      {"dict", 1, 0, 'd'},
      {0, 0, 0, 0}
    };

  bool setStart,setEnd,setMulti;
  int start,end;
  string dictFileName,solFileName;
  VFile dictFile,solFile;
  string tag1,tag2;
  Dict d1,d2;

  cout<<"* Print-solution v"<<PRINTSOL_VERSION<<" by Benjamin Han"<<endl<<endl;

  // ------------------------------------------------------------
  // Parsing command-line options
  // ------------------------------------------------------------

  setStart=setEnd=setMulti=false;
  start=end=1;
  dictFileName=DICTFILE;

  opterr=0;   // turn off the default getopt() error messages
  
  while ((opt=
	  getopt_long_only (argc, argv, "s:e:md:",longOpts, &longOptID))!=-1)
    switch (opt)
      {
      case 's':
	if ((start=atoi(optarg))<=0)
	  {
	    printHelp();
	    cerr<<"* Negative or zero starting index."<<endl;
	    return ERR_ILLEGAL_START;
	  }

	if (setEnd)
	  {
	    if (end<start)
	      {
		printHelp();
		cerr<<"* Illegal range."<<endl;
		return ERR_ILLEGAL_RANGE;
	      }
	  }
	else end=start;

	setStart=true;
	break;
	
      case 'e':
	if ((end=atoi(optarg))<=0)
	  {
	    printHelp();
	    cerr<<"* Negative or zero ending index."<<endl;
	    return ERR_ILLEGAL_END;
	  }

	if (setStart)
	  {
	    if (end<start)
	      {
		printHelp();
		cerr<<"* Illegal range."<<endl;
		return ERR_ILLEGAL_RANGE;
	      }
	  }
	
	setEnd=true;
	break;

      case 'm':
	setMulti=true;
	break;

      case 'd':
	dictFileName=optarg;
	break;
	
      case '?':
	printHelp();
	cerr<<"* Option "<<argv[optind-1]
	    <<" is illegal or missing an argument."<<endl;
	return ERR_ILLEGAL_OPT;
	break;
	
      default:
	cerr<<"* Internal command-line parsing error, code = 0"<<oct<<opt<<'.'<<endl;
	return ERR_INTERNAL_CMDLINE;
      }

  if (optind!=argc-1)
    {
      printHelp();
      cerr<<"* Missing a solution file name or too many file names."<<endl;
      return ERR_NO_INPUT_FILE;
    }
  else solFileName=argv[optind];

  // ------------------------------------------------------------
  // Printing solutions
  // ------------------------------------------------------------  

  try
    {
      dictFile.open(dictFileName,VFILE_READ);
      solFile.open(solFileName,VFILE_READ);

      dictFile>>tag1;
      solFile>>tag2;
      if (tag1!=tag2) throw SolFileMismatch("");

      cout<<"* Reading the dictionary file "<<dictFileName<<" ..."<<flush;
      dictFile>>numIter>>c1>>c2>>d1>>d2;

      dictFile.close();
      d1.setReadMode();
      d2.setReadMode();

      cout<<"done."<<endl<<endl;
      
      printStat(d1,d2);

      printSol(start,end,setMulti,d1,d2,solFile);
    }
  catch (Exception &e)
    {
      cerr<<"* ERROR: "<<e<<endl;
      return ERR_OPEN_INPUT_FILE;
    }
  
  return ERR_SUCCESS;
}
