// ======================================================================
// readdic.cpp - read/generate a dictionary file
// 
// 111200: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "analyzer.hpp"

void DictEntryToConstraints::operator () (int wid, const DictEntry &de)
{
  int i;
  int s=de.cVec.size();
  CVec::const_iterator ci;
  Var &v=cons[id];

  v.vid=id++;
  v.r.resize(s);
  for (i=0,ci=de.cVec.begin();i<s;++i,++ci)
    {
      // yeah I know Concept and Candidate are two homomorphic structures
      // but the idea is to make dict and solver modules independent from
      // each other
      v.r[i].val=ci->wordID;
      v.r[i].p=ci->c;
    }
}

void DictEntryToBestSol::operator () (int wid, const DictEntry &de)
{
  bestInitSol[wid]=max_element(de.cVec.begin(),de.cVec.end(),
			       conceptConfAscendingOrder)->wordID;
}

int readDictionary (Settings &settings, Dict &d1, Dict &d2)
{
  fstream corpusFile,postFile;
  string sentence,tags,str;
  int i,id=0;
  VFile dictFile;

  try
    {
      dictFile.open(DICTFILE,VFILE_READ);
      cout<<"* "<<DICTFILE<<" already exists, trying to read it in... "<<flush;

      dictFile>>str;
      if (settings.corpusFileName!=str)
	{
	  cout<<endl<<"* "<<DICTFILE<<" was generated from "<<str<<", not "
	      <<settings.corpusFileName<<'.'<<endl;
	  cout<<"* Trying to regenerate it..."<<endl;
	  throw Exception("","");
	}

      dictFile>>settings.numIter>>settings.c1>>settings.c2>>d1>>d2;
      dictFile.close();
      cout<<"done."<<endl<<endl;

      cout<<"* "<<DICTFILE<<" was generated using the following parameters:"<<endl;
      cout<<"* cutoff 1   = "<<settings.c1<<endl;
      cout<<"* cutoff 2   = "<<settings.c2<<endl;
      cout<<"* iterations = "<<settings.numIter<<endl<<endl;

      printStat(d1,d2);
    }

  catch (Exception &e)
    {
      corpusFile.open(settings.corpusFileName.c_str(),ios::in);
      if (!corpusFile)
	{
	  cerr<<"* ERROR: Cannot open file "<<settings.corpusFileName<<" for reading."<<endl;
	  return ERR_OPEN_INPUT_FILE;
	}

      postFile.open(settings.postFileName.c_str(),ios::in);
      if (!postFile)
	{
	  cerr<<"* ERROR: Cannot open file "<<settings.postFileName<<" for reading."<<endl;
	  return ERR_OPEN_INPUT_FILE;
	}

      for (i=0;getline(corpusFile,sentence);)
	{
	  // only put the first sentence of a sentence pair into sList
	  processString(sentence);
	  if (sentence.length()==0) i=0;
	  else
	    switch (i)
	      {
	      case 0:  // L2
		i=1;
		if (settings.setBackward) d2.addSentence(id,sentence);
		else 
		  {
		    // L2 is the target language
		    if (!getline(postFile,tags))
		      {
			cerr<<"* ERROR: POST-file doesn't match the corpus-file."<<endl;
			return ERR_POST_FILE_MISMATCH;
		      }

		    try { d2.addSentence(id,sentence,tags); }
		    catch (Exception &e)
		      {
			cerr<<"* ERROR: "<<e<<endl;
			return ERR_TAGS_PARSING;
		      }
		  }
		break;
		
	      case 1:  // L1
		i=2;
		if (!settings.setBackward) d1.addSentence(id,sentence);
		else 
		  {
		    // L1 is the target language
		    if (!getline(postFile,tags))
		      {
			cerr<<"* ERROR: POST-file doesn't match the corpus-file."<<endl;
			return ERR_POST_FILE_MISMATCH;
		      }

		    try { d1.addSentence(id,sentence,tags); }
		    catch (Exception &e)
		      {
			cerr<<"* ERROR: "<<e<<endl;
			return ERR_TAGS_PARSING;
		      }
		  }
		++id;
		break;

	      case 2:
		cerr<<"* ERROR: More then 2 consecutive sentences."<<endl;
		return ERR_ILLEGAL_CORPUS_FILE_FORMAT;	
		
	      default:
		break;
	      }
	}
      
      corpusFile.close();

      cout<<"* cutoff 1   = "<<settings.c1<<endl;
      cout<<"* cutoff 2   = "<<settings.c2<<endl;
      cout<<"* iterations = "<<settings.numIter<<endl<<endl;

      printStat(d1,d2);

      cout<<"* Computing candidate sets... "<<flush;
      d1.computeConcepts(d2,settings.c1,settings.c2,settings.numIter);
      cout<<"done."<<endl;
      
      cout<<"* Writing to file "<<DICTFILE<<"... "<<flush;
      try { dictFile.open(DICTFILE,VFILE_WRITE); }
      catch (Exception &e)
	{
	  cerr<<"* ERROR: "<<e<<endl;
	  return ERR_WRITING_DICTFILE;
	}

      dictFile<<settings.corpusFileName<<settings.numIter<<settings.c1<<settings.c2<<d1<<d2;

      dictFile.close();
      cout<<"done."<<endl<<endl;
    }

  return ERR_SUCCESS;
}
