// ======================================================================
// settings.cpp - parsing the command-line arguments for Analyzer
// 
// 110900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 021701: Benjamin Han <benhdj@cs.cmu.edu> Add '-addBest' command-line
//         option.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <getopt.h>
#include <limits.h>
#include <float.h>

#include "analyzer.hpp"

int Settings::parse (int argc, char *argv[])
{
  int i,cutoff,numSetOpRate;
  double sum;
  bool setOpRate[GA_OP_NUM];

  int opt;
  int longOptID = 0;
  static struct option longOpts[] =
    {
      {"backward", 1, 0, 'b'},
      {"c1", 1, 0, 0},
      {"c2", 1, 0, 0},
      {"relaxed", 1, 0, 'r'},
      {"addBest", 1, 0, 'a'},
      {"iter", 1, 0, 'n'},
      {"solFile", 1, 0, 's'},
      {"popsize", 1, 0, 'p'},
      {"goal", 1, 0, 'g'},
      {"epochs", 1, 0, 'e'},
      {"crossover", 1, 0, 'x'},
      {"mutate", 1, 0, 'm'},
      {"creep", 1, 0, 'c'},
      {"history", 1, 0, 'h'},
      {"printInt", 1, 0, 'i'},
      {"display", 1, 0, 'd'},
      {0, 0, 0, 0}
    };

  opterr=0;   // turn off the default getopt() error messages

  sum=0.0;
  numSetOpRate=0;
  for (i=0;i<GA_OP_NUM;++i) setOpRate[i]=false;
  
  while ((opt=
	  getopt_long_only (argc, argv, "br:an:sp:e:g:x:m:c:i:d",longOpts, &longOptID))!=-1)
    switch (opt)
      {
      case 'b':
	setBackward=true;
	break;

      case 0:
	if ((cutoff=atoi(optarg))==0)
	  {
	    printHelp();
	    cerr<<"* Zero cutoff value."<<endl;
	    return ERR_ZERO_CUTOFF;
	  }

	if (longOptID==1) c1=cutoff;
	else c2=cutoff;

	break;

      case 'r':
	if ((relaxedMode=atoi(optarg))!=1 && relaxedMode!=2)
	  {
	    printHelp();
	    cerr<<"* Relaxed mode is neither 1 nor 2."<<endl;
	    return ERR_ILLEGAL_RELAXED_MODE;
	  }

	break;

      case 'a':
	setAddBest=true;
	break;

      case 'n':
	if ((numIter=atoi(optarg))<1)
	  {
	    printHelp();
	    cerr<<"* # of iterations in computing the candidate concepts must be >=1."<<endl;
	    return ERR_ILLEGAL_NUM_ITER;
	  }

	break;

      case 's':
	setSolFile=true;
	break;

      case 'p':
	if ((popSize=atoi(optarg))<=0)
	  {
	    printHelp();
	    cerr<<"* Negative or zero population size."<<endl;
	    return ERR_ZERO_POPSIZE;
	  }

	break;
	
      case 'e':
	maxEpoch=atoi(optarg);
	break;

      case 'g':
	goalScore=atof(optarg);
	break;
	
      case 'x':
	if ((accOpRate[GA_OP_CROSSOVER]=atof(optarg))<0.0 || 
	    accOpRate[GA_OP_CROSSOVER]>1.0)
	  {
	    printHelp();
	    cerr<<"* Illegal operator rate. (should be in [0,1])"<<endl;
	    return ERR_ILLEGAL_OP_RATE;
	  }

	setOpRate[GA_OP_CROSSOVER]=true;
	sum+=accOpRate[GA_OP_CROSSOVER];
	++numSetOpRate;
	break;
	
      case 'm':
	if ((accOpRate[GA_OP_MUTATE]=atof(optarg))<0.0 || 
	    accOpRate[GA_OP_MUTATE]>1.0)
	  {
	    printHelp();
	    cerr<<"* Illegal operator rate. (should be in [0,1])"<<endl;
	    return ERR_ILLEGAL_OP_RATE;
	  }

	setOpRate[GA_OP_MUTATE]=true;
	sum+=accOpRate[GA_OP_MUTATE];
	++numSetOpRate;
	break;

      case 'c':
	if ((accOpRate[GA_OP_CREEP]=atof(optarg))<0.0 || 
	    accOpRate[GA_OP_CREEP]>1.0)
	  {
	    printHelp();
	    cerr<<"* Illegal operator rate. (should be in [0,1])"<<endl;
	    return ERR_ILLEGAL_OP_RATE;
	  }

	setOpRate[GA_OP_CREEP]=true;
	sum+=accOpRate[GA_OP_CREEP];
	++numSetOpRate;
	break;

      case 'h':
	if ((opHistorySize=atoi(optarg))<=0)
	  {
	    printHelp();
	    cerr<<"* Negative operator history length."<<endl;
	    return ERR_ILLEGAL_OP_HPPISTORY_SIZE;
	  }
	
	break;

      case 'i':
	if ((printInt=atoi(optarg))<=0)
	  {
	    printHelp();
	    cerr<<"* Negative or zero printing interval."<<endl;
	    return ERR_ILLEGAL_PRTINT;
	  }
	
	break;

      case 'd':
	setDisplay=true;
	break;
	
      case '?':
	printHelp();
	cerr<<"* Option "<<argv[optind-1]
	    <<" is illegal or missing an argument."<<endl;
	return ERR_ILLEGAL_OPT;
	break;
	
      default:
	cerr<<"* Internal command-line parsing error, code = 0"<<oct<<opt<<'.'<<endl;
	return ERR_INTERNAL_CMDLINE;
      }

  if (optind!=argc-2)
    {
      printHelp();
      cerr<<"* Missing corpus file/POST file or too many file names are given."<<endl;
      return ERR_NO_INPUT_FILE;
    }
  else 
    {
      corpusFileName=argv[optind];
      postFileName=argv[optind+1];
    }  

  // check operator rate settings
  if (sum>1.0 || (numSetOpRate==GA_OP_NUM && sum<1.0))
    {
      printHelp();
      cerr<<"* Sum of the operator rates is not 1.0."<<endl;
      return ERR_ILLEGAL_OP_RATE;
    }
  
  if (numSetOpRate<GA_OP_NUM)
    {
      sum=(1.0-sum)/(GA_OP_NUM-numSetOpRate);
      for (i=0;i<GA_OP_NUM;++i)
	if (!setOpRate[i]) accOpRate[i]=sum;
    }
  
  if (relaxedMode==0 && accOpRate[GA_OP_CREEP]!=0.0)
    {
      printHelp();
      cerr<<"* GA operator Creep is not available under non-relaxed mode."<<endl;
      return ERR_OP_NO_CREEP;
    }
  
  for (i=1;i<GA_OP_NUM;++i) accOpRate[i]+=accOpRate[i-1];
  
  if (maxEpoch<0) maxEpoch=INT_MAX;
  if (goalScore<0.0) goalScore=DBL_MAX;
  if (relaxedMode==0 && setAddBest)
    {
      printHelp();
      cerr<<"* Cannot set addBest under non-relaxed mode."<<endl;
      return ERR_NO_ADDBEST;
    }

  if (setAddBest && setSolFile)
    {
      printHelp();
      cerr<<"* Cannot set addBest and solFile at the same time."<<endl;
      return ERR_NO_ADDBEST_SOLFILE;
    }

  return ERR_SUCCESS;
}
