// ======================================================================
// soltest.cpp - demo code for solver
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102300: Benjamin Han <benhdj@cs.cmu.edu> Added a dummy CreepFunc.
// 021701: Benjamin Han <benhdj@cs.cmu.edu> set addBest=true when calling
//         Solver::init().
// 080501: Benjamin Han <benhdj@cs.cmu.edu> Bugs: the probability of each
//         candidate is 1.0/k, not 1/k (which evaluates to 0); added a few
//         clarification comments; turn off 'add-best' in calling 
//         Solver::init().
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <fstream>

#include <math.h>

#include "solver.hpp"

#define SOLSIZE          10
#define MAX_RSIZE        10      // max range set size
#define MAX_VAL          20
#define RELAXED_MODE     0       // no relaxation
#define POPSIZE          50
#define OP_HPPISTORY_SIZE  3
#define PRT_INT          50
#define MAX_EPOCH           1000

class CreepFunc: public SolCreepFunc
{
public:
  bool operator () (Solution &sol)
  {
    // for demo: do nothing
    return false;
  }
};

class EvalFunc: public SolEvalFunc
{
public:
  double operator () (const Solution &sol);
};

double EvalFunc::operator () (const Solution &sol)
{
  Constraints::size_type i;
  double score;

  for (i=0,score=0.0;i<sol.size();++i)
    if (sol[i]!=UNASSIGNED) score+=sol[i];
  return score;
}

int main ()
{
  int i,j,k,val;
  RangeSet rs;
  RangeSet::const_iterator ri;
  Constraints cons;
  Solver solver;
  PDist accOpRate(GA_OP_NUM);
  double maxScore;
  CreepFunc cf;
  EvalFunc ef;
  fstream statFile("soltest.stat",ios::out);

  // generate the constraint for each solution element
  // the goal is to find a set of binary-exclusive solution elements from
  // the possible candidates (each element has its own candidates in a Range
  // structure with the probability) such that the sum of the solution elements
  // is maximal
  cons.resize(SOLSIZE);
  for (i=0,maxScore=0.0;i<SOLSIZE;++i,rs.clear())
    {
      cons[i].vid=i;
      cons[i].r.resize(k=random()%MAX_RSIZE+1);
      for (j=0;j<k;++j)
	{
	  do
	    {
	      val=random()%MAX_VAL;
	    } while (rs.find(val)!=rs.end());
	  rs.insert(val);
	}
      for (j=0,ri=rs.begin();ri!=rs.end();++j,++ri)
	{
	  cons[i].r[j].val=*ri;
	  cons[i].r[j].p=1.0/k;
	}

      maxScore+=cons[i].r[j-1].val;

      cout<<"* "<<i<<":";
      for (Range::const_iterator ri=cons[i].r.begin();ri!=cons[i].r.end();
	   ++ri)
	cout<<ri->val<<' ';
      cout<<endl;
    }
  // over-optimisitic because some choices conflict with another (the same 
  // ones)
  cout<<endl<<"* max score (over-optimistic) = "<<maxScore<<endl<<endl;
  
  accOpRate[GA_OP_CROSSOVER]=.5;
  accOpRate[GA_OP_MUTATE]=1.0;

  // do not "add best" since each candidate solution is equally probable
  solver.init(cons,RELAXED_MODE,POPSIZE,accOpRate,OP_HPPISTORY_SIZE,cf,ef,PRT_INT,statFile,false);
  solver.printStat();

  cout<<endl<<"* Solver running until "<<MAX_EPOCH<<" epochs or the best score reaches "<<maxScore<<"..."<<flush;
  solver.run(MAX_EPOCH,maxScore);
  cout<<"done."<<endl<<endl;
  solver.printStat();

  cout<<endl<<"* Best solution: ";
  cout<<solver.getBestSolution()<<endl<<endl;

  return 0;
}
