// ======================================================================
// solver.cpp - a solver for optimization constraint satisfaction problem
//             using a genetic algorithm
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Added GA operator Creep 
//         (local optimization) - it's problem specific and needed to be
//         'hooked' from outside.
// 110500: Benjamin Han <benhdj@cs.cmu.edu> Support for incremental
//         evaluation based on the changed assignments in a solution is
//         in place
// 110900: Benjamin Han <benhdj@cs.cmu.edu> New Solver::init() added for
//         initializing a population by loading a solution file.
// 111200: Benjamin Han <benhdj@cs.cmu.edu> Change the behavior of
//         crossover - now if no novelty is introduced mutation will take
//         over
// 021501: Benjamin Han <benhdj@cs.cmu.edu> Minor fix: redundant
//         "Population size" printout in analysis.stat.
// 021701: Benjamin Han <benhdj@cs.cmu.edu> Add 'addBest' argument to 
//         Solver::init().
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <algorithm>
#include <map>

#include <float.h>

#include "solver.hpp"

#define C       1.2  // constant for scaling fitness values, usually [1.2,2.0]
#define F_MIN   0.1  // min converted fitness when linear scaling case 1 fails
#define P_COEFF 0.1  // the portion of the operator credit propagating back

void Constraints::randomCopy (const Constraints &src)
{
  RandomIntGenerator rg;
  *this=src;
  random_shuffle(begin(),end(),rg);
}

int Constraints::satisfy (Solution &sol)
{
  Range range;
  Range::const_iterator ri;
  Range::size_type size;
  RangeSet &assigned=sol.assigned;
  iterator ci;

  int numSVars=0;       // # of vars satisfied

  // Sequential assignment has O(n^2) complexity, faster than selecting the
  // var with singleton range first, which has complexity O(n^2 * log n).
  // Since it's already in random order I think it makes no impact to the GA
  // performance.
  for (ci=begin();ci!=end();++ci,range.clear())
    {
      const Range &r=ci->r;

      // range = r - assigned
      // TO-DO: can be further sped up by only generating partial set (up to
      //        the select(range)-th element)
      for (ri=r.begin();ri!=r.end();++ri)
	if (assigned.find(ri->val)==assigned.end()) range.push_back(*ri);

      if ((size=range.size()) == 0) continue;
      else
	{
	  sol.changed.insert(ci->vid);
	  ++numSVars;
	  if (size==1) assigned.insert(sol[ci->vid]=range.begin()->val);
	  else assigned.insert(sol[ci->vid]=select(range));
	}
    }

  sol.numSVars+=numSVars;

  assert(sol.numSVars==assigned.size());
  return numSVars;
}

int Constraints::satisfyRelaxed (Solution &sol)
{
  Range range;
  iterator ci;

  int numSVars=0;       // # of vars satisfied

  // Sequential assignment has O(n^2) complexity, faster than selecting the
  // var with singleton range first, which has complexity O(n^2 * log n).
  // Since it's already in random order I think it makes no impact to the GA
  // performance.
  for (ci=begin();ci!=end();++ci)
    {
      sol.changed.insert(ci->vid);
      sol[ci->vid]=select(ci->r);  // no longer do we need to maintain 'assigned'
      ++numSVars;
    }

  sol.numSVars+=numSVars;

  return numSVars;
}

// define static members of Solution
#ifndef NDEBUG
bool Solution::setRelaxedModeF=false;
bool Solution::setOpHistorySizeF=false;
bool Solution::setConstraintsF=false;
bool Solution::setCreepFuncPtrF=false;
bool Solution::setEvalFuncPtrF=false;
#endif

int Solution::relaxedMode=0;
list<int>::size_type Solution::opHistorySize=0;
Constraints Solution::c0;
SolCreepFunc *Solution::creepFuncPtr=NULL;
SolEvalFunc *Solution::evalFuncPtr=NULL;

// TO-DO: can be sped up by taking non-symmetry into consideration
void Solution::satisfy (size_type cp)
{
  typedef map<int,size_type> VMap;  // map of (val,id) pairs

  size_type i,s=size();
  VMap vMap;
  VMap::const_iterator vi;
  iterator si;
  Var v;
  Constraints cons;
  Constraints::iterator ci;
  double p1,p2;

  assert(relaxedMode==0);  // only non-relaxed GA operators can call this

  // the solution after crossover: need to count numSVars and
  // construct 'assigned' again
  numSVars=0;
  assigned.clear();  

  // building a var assignment map
  for (i=0,si=begin();i<cp;++i,++si)
    if (*si!=UNASSIGNED)
      {
	vMap.insert(make_pair(*si,i));
	assigned.insert(*si);
	++numSVars;
      }
    else
      {
	// originally unassigned vars are pushed into cons
	v.vid=i;
	cons.push_back(v);
      }

  // check for conflicts in the second half of the solution against vMap
  for (;i<s;++i,++si)
    if (*si!=UNASSIGNED)
      if ((vi=vMap.find(*si))!=vMap.end())
	{
	  // conflict arises - keep the superior one and put the inferior one
	  // into 'cons', waiting for being satisfied

	  p1=lower_bound(c0[vi->second].r.begin(),c0[vi->second].r.end(),*si)->p;
	  p2=lower_bound(c0[i].r.begin(),c0[i].r.end(),*si)->p;

	  if (p1>=p2) (*this)[v.vid=i]=UNASSIGNED;   // the current one is inferior
	  else (*this)[v.vid=vi->second]=UNASSIGNED; // the current one is superior

	  changed.insert(v.vid);
	  cons.push_back(v);
	}
      else
	{
	  ++numSVars;
	  assigned.insert(*si);
	}
    else
      {
	// originally unassigned vars are pushed into cons
	v.vid=i;
	cons.push_back(v);
      }

  // building ranges for 'cons'
  for (ci=cons.begin();ci!=cons.end();++ci) ci->r=c0[ci->vid].r;

  assert(numSVars==assigned.size());
  cons.satisfy(*this);
}

void Solution::assign (const Solution &sol)
{
  iterator si1;
  const_iterator si2;
  Solution::size_type i;
  
  if (relaxedMode==0) assigned=sol.assigned;
  numSVars=sol.numSVars;
  score=sol.score;
  opHistory=sol.opHistory;
  
  // solution assignment - update 'changed' also!
  for (si1=begin(),si2=sol.begin(),i=0;si1!=end();++si1,++si2,++i)
    if (*si1!=*si2)
      {
	*si1=*si2;
	changed.insert(i);
      }
}

double Solution::evaluate (double bestScore, PDist &accOpRates)
{ 
  double delta,coeff;
  int i;
  list<int>::const_iterator hi;

  assert(setOpHistorySizeF && setEvalFuncPtrF);
  score=(*evalFuncPtr)(*this);

  assert(score>=0.0);

  changed.clear();

  if (score>bestScore)
    {
      // compute the non-accumulated dist.
      for (i=GA_OP_NUM-1;i>=1;i--)
	accOpRates[i]-=accOpRates[i-1];

      delta=(score-bestScore)/bestScore;
      coeff=1.0;
      for (i=0,hi=opHistory.begin();hi!=opHistory.end();++i,++hi)
	{
	  delta*=coeff;
	  coeff*=P_COEFF;
	  accOpRates[*hi]+=delta;
	}

      // restore the accumulated dist.
      for (i=1;i<GA_OP_NUM;++i) accOpRates[i]+=accOpRates[i-1];
      for (i=0;i<GA_OP_NUM;++i) accOpRates[i]/=accOpRates[GA_OP_NUM-1];
    }
  return score; 
}
  
bool Solution::crossover (const Solution &sol, Solution &child1, Solution &child2) const
{
  size_type i,s=size();   // s must be greater than 1
  size_type cp=getRandInt(s-1)+1;
  const_iterator si1,si2;
  RangeSet cs;            // changed set
  RangeSet::const_iterator ri;

  assert(setRelaxedModeF && setOpHistorySizeF && setConstraintsF);

  // vMap records the first assignment of a specific value
  
  //  v0 v1 ...  v_{cp-1} ... v_{s-1}
  //    ^                ^   ^        
  //    1   ...         cp  s-1

  // check if the new solution is any different from the parents
  if (cp<s-cp)
    for (si1=begin(),si2=sol.begin(),i=0;i<cp;++si1,++si2,++i)
      if (*si1!=*si2) cs.insert(i);
  else
    for (si1=begin()+cp,si2=sol.begin()+cp,i=cp;i<size();++si1,++si2,++i)
      if (*si1!=*si2) cs.insert(i);

  // if no novelty is introduced the operation fails - we don't even
  // bother updating the operator history because the operator makes
  // zero contribution
  if (cs.size()==0) return false;

  child1.assign(*this);
  child2.assign(sol);

  // record operator history
  child1.opHistory.push_back(GA_OP_CROSSOVER);
  if (child1.opHistory.size()>opHistorySize) 
    child1.opHistory.erase(child1.opHistory.begin());
  child2.opHistory.push_back(GA_OP_CROSSOVER);
  if (child2.opHistory.size()>opHistorySize)
    child2.opHistory.erase(child2.opHistory.begin());

  for (ri=cs.begin();ri!=cs.end();++ri)
    {
      std::swap(child1[*ri],child2[*ri]);
      child1.changed.insert(*ri);
      child2.changed.insert(*ri);
    }

  if (relaxedMode!=0)
    {
      Constraints cons1,cons2;
      Var v;

      // re-count # of satisfied vars
      child1.numSVars=child2.numSVars=0;
      for (i=0;i<s;++i)
	if (child1[i]==UNASSIGNED && child2[i]==UNASSIGNED)
	  {
	    v.vid=i;
	    v.r=c0[i].r;
	    cons1.push_back(v);
	    cons2.push_back(v);
	  }
	else if (child1[i]!=UNASSIGNED && child2[i]!=UNASSIGNED)
	  {
	    ++child1.numSVars;
	    ++child2.numSVars;
	  }
	else
	  {
	    v.vid=i;
	    v.r=c0[i].r;
	    if (child1[i]!=UNASSIGNED)
	      {
		++child1.numSVars;
		cons2.push_back(v);
	      }
	    else
	      {
		++child2.numSVars;
		cons1.push_back(v);
	      }
	  }

      // try to assign all unassigned vars
      if (cons1.size()) cons1.satisfyRelaxed(child1);
      if (cons2.size()) cons2.satisfyRelaxed(child2);
    }
  else
    {
      // find conflicts in the offsprings and try to satisfy the constraints
      child1.satisfy(cp);
      child2.satisfy(cp);
    }

  return true;
}

void Solution::mutate ()
{
  size_type i,s=size();       // s must be greater than 1
  size_type mp=getRandInt(s);     // mutation point
  int newVal;
  Range r;
  const Range &rmp=c0[mp].r;

  assert(setRelaxedModeF && setOpHistorySizeF && setConstraintsF);

  // record operator history
  opHistory.push_back(GA_OP_MUTATE);
  if (opHistory.size()>opHistorySize) opHistory.erase(opHistory.begin());

  changed.insert(mp);

  if (relaxedMode!=0)
    {
      if ((*this)[mp]!=UNASSIGNED)
	if (rmp.size()>1) (*this)[mp]=select(rmp,(*this)[mp]);
	else
	  {
	    (*this)[mp]=UNASSIGNED;
	    numSVars--;
	  }
      else
	{
	  if (rmp.size()>1) (*this)[mp]=select(rmp);
	  else (*this)[mp]=rmp.begin()->val;
	  ++numSVars;
	}

      return;
    }

  if ((*this)[mp]!=UNASSIGNED)
    // guaranteed that a different value from the old one will be
    // randomly picked
    if (rmp.size()!=1) 
      {
	assigned.erase(assigned.find((*this)[mp]));  // unregister the old val
	// it's ok if 'assigned' already has the value - it's a set
	assigned.insert((*this)[mp]=(newVal=select(rmp,(*this)[mp])));
      }
    else
      {
	assigned.erase(assigned.find((*this)[mp]));
	(*this)[mp]=UNASSIGNED;	
	numSVars--;

	assert(numSVars==assigned.size());

	return;
      }
  else  // rmp must not be empty
    { 
      // it's ok if 'assigned' already has the value - it's a set
      assigned.insert((*this)[mp]=(newVal=select(rmp)));
      ++numSVars;
    }

  // check for conflict (at most one is possible)
  for (i=0;i<s;++i) if ((*this)[i]==newVal && i!=mp) break;

  // i now points to the conflict point
  if (i<s)
    {
      changed.insert(i);
      numSVars--;

      // r = range - assigned
      // (range = c0[i].r)
      const Range &range=c0[i].r;
      Range::const_iterator ri;

      for (ri=range.begin();ri!=range.end();++ri)
	if (assigned.find(ri->val)==assigned.end()) r.push_back(*ri);

      // reassign conflicting var
      if (r.size())
	{
	  assigned.insert((*this)[i]=select(r));
	  ++numSVars;
	}
      else (*this)[i]=UNASSIGNED;  // no possible value available
    }

  assert(numSVars==assigned.size());
}

bool Solution::creep ()
{
  assert(setRelaxedModeF && setOpHistorySizeF && setConstraintsF 
	 && setCreepFuncPtrF);

  // record operator history
  opHistory.push_back(GA_OP_CREEP);
  if (opHistory.size()>opHistorySize) opHistory.erase(opHistory.begin());

  return (*creepFuncPtr)(*this);
}

void Solution::readFromVFile (VFile &vf)
{
  vf>>numSVars>>score>>*((vector<int>*)this)>>opHistory;
  // WARNING: 'assigned' is not reconstructed!
}

void Solution::writeToVFile (VFile &vf) const
{
  vf<<numSVars<<score<<*((vector<int>*)this)<<opHistory;
  // WARNING 'assigned' is not saved
}

ostream &operator << (ostream &s, const Solution &sol)
{
  copy(sol.begin(),sol.end(),ostream_iterator<int>(s," "));
  return s;
}

void Solver::computeFitness ()
{
  vector<Solution>::size_type i,s=population.size();
  double score,accScore;

  bestID=worstID1=worstID2=0;       // in the unlikely situation we may have
  bestScore=DBL_MIN;                //  all scores as DBL_MAX or all scores
  worstScore1=worstScore2=DBL_MAX;  //  as DBL_MIN
    
  for (i=0,accScore=0.0;i<s;++i)
    {
      Solution &sol=population[i];

      score=sol.getScore();

      if (bestScore<score)
	{
	  bestScore=score;
	  bestID=i;
	}

      // invariant: worstScore1<=worstScore2 && worstID1 != worstID2
      if (worstScore1>score)
	{
	  worstScore2=worstScore1;
	  worstID2=worstID1;
	  worstScore1=score;
	  worstID1=i;
	}

      // score is not the lowest one - worstScore2 may be updated
      else if (worstScore2>score)
	{
	  worstScore2=score;
	  worstID2=i;
	}
      
      accScore+=score;
    }
  avgScore=accScore/s;

  // solving for linear scaling coefficients [Goldberg 89]; C and F_MIN
  // are constants: (see the code for the decision process)
  // 1. f_avg     = a * f_avg + b  => b = (1-a) * f_avg
  //    C * f_avg = a * f_max + b  => a = (C-1) * f_avg / (f_max - f_avg)
  // 2. f_avg     = a * f_avg + b
  //    F_MIN     = a * f_min + b  => a = (f_avg - F_MIN)/(f_avg - f_min)
  // 3. a = 1.0 and b = 0.0 (scaling failed)

  scaleFail=true;
  if (bestScore!=avgScore)
    {
      // case 1
      a=(C-1)*avgScore/(bestScore-avgScore);
      b=(1-a)*avgScore;
      if (a*worstScore1+b>=F_MIN) scaleFail=false;
    }

  if (scaleFail)
    if (worstScore1!=avgScore)
      {
	// case 2
	a=(avgScore-F_MIN)/(avgScore-worstScore1);
	b=(1-a)*avgScore;
	scaleFail=false;
      }

  // scaling if scaleFail == false
  if (scaleFail) 
    for (i=1,accP1[0]=population[0].getScore();i<s;++i)
      accP1[i]=accP1[i-1]+population[i].getScore();
  else
    for (i=1,accP1[0]=a*population[0].getScore()+b;i<s;++i)
      accP1[i]=accP1[i-1]+a*population[i].getScore()+b;
}

Solver::Solver ():statFilePtr(NULL)
{
  #ifndef NDEBUG
  initFlag=false;
  #endif

  initRandom();
}

void Solver::init (const Constraints &c0, int relaxedMode, list<int>::size_type opHistorySize,
		   SolCreepFunc &creepFunc, SolEvalFunc &evalFunc, 
		   int prtInt, fstream &statFile, VFile &solFile)
{
  Constraints::size_type i;
  RangeSet changed;

  printHeader=false;

  op=-1;

  Solution::setRelaxedMode(relaxedMode);
  Solution::setOpHistorySize(opHistorySize);
  Solution::setConstraints(c0);
  Solution::setCreepFunc(creepFunc);
  Solution::setEvalFunc(evalFunc);

  numEpochs=0;
  Solver::prtInt=prtInt;
  statFilePtr=&statFile;

  solSize=c0.size();
  solFile>>population>>accP2;
  accP1.resize(population.size());

  assert(solSize>1);   // solution size<=1 is meaningless!

  for (i=0;i<solSize;++i) changed.insert(i);

  // initialize the population
  for (i=0;i<population.size();++i)
    {
      Solution &sol=population[i];

      sol.setID(i);

      sol.changed=changed;          // force full evaluation
      sol.evaluate(DBL_MAX,accP2);  // prevent operator adaptation in the 1st run
    }

  computeFitness();

  #ifndef NDEBUG
  initFlag=true;
  #endif
}

void Solver::init (const Constraints &c0, int relaxedMode, size_type popSize,
		   const PDist &accOpRates, list<int>::size_type opHistorySize,
		   SolCreepFunc &creepFunc, SolEvalFunc &evalFunc, 
		   int prtInt, fstream &statFile, bool addBest)
{
  Constraints::size_type i,j;
  int k,l,m=0;  // m=0 just to suppress the warning of "use before initialized"
  Range::const_iterator ri;
  double pMax;
  Constraints cons;
  RangeSet changed;

  printHeader=true;

  op=-1;

  Solution::setRelaxedMode(relaxedMode);
  Solution::setOpHistorySize(opHistorySize);
  Solution::setConstraints(c0);
  Solution::setCreepFunc(creepFunc);
  Solution::setEvalFunc(evalFunc);

  numEpochs=0;
  Solver::prtInt=prtInt;
  statFilePtr=&statFile;

  solSize=c0.size();
  population.resize(popSize);
  accP1.resize(popSize);

  assert(solSize>1);   // solution size<=1 is meaningless!

  accP2=accOpRates;
  for (i=0;i<solSize;++i) changed.insert(i);

  // initialize the population
  for (i=0;i<popSize;++i)
    {
      Solution &sol=population[i];

      sol.resize(solSize);
      sol.setID(i);

      if (i || !addBest)
	{
	  fill(sol.begin(),sol.end(),UNASSIGNED);
	  cons.randomCopy(c0);

	  if (relaxedMode==1) cons.satisfyRelaxed(sol);
	  else cons.satisfy(sol);
	}
      else
	{
	  for (j=0;j<solSize;++j)
	    {
	      l=c0[j].r.size();
	      for (ri=c0[j].r.begin(),pMax=0.0,k=0;k<l;++ri,++k)
		if (pMax<ri->p)
		  {
		    m=k;
		    pMax=ri->p;
		  }

	      assert(pMax>0.0);
	      sol[j]=c0[j].r[m].val;
	    }

	  sol.allSatisfied();
	}

      sol.changed=changed;          // force full evaluation
      sol.evaluate(DBL_MAX,accP2);  // prevent operator adaptation in the 1st run
    }

  computeFitness();

  #ifndef NDEBUG
  initFlag=true;
  #endif
}

void Solver::run (int maxEpoch, double goalScore)
{
  assert(initFlag==true);

  int i,s1,s2;
  Solution sol;

  assert(maxEpoch>0 && goalScore>0.0);

  if (printHeader)
    {
      (*statFilePtr)<<"* Solution size = "<<solSize<<endl;
      (*statFilePtr)<<"* Population size = "<<population.size()<<endl;
      (*statFilePtr)<<"* Operator history size  = "<<Solution::getOpHistorySize()<<endl;
      (*statFilePtr)<<"* Max epoch = "<<maxEpoch<<endl;
      (*statFilePtr)<<"* Goal score = "<<goalScore<<endl;
      (*statFilePtr)<<"* Print interval = "<<prtInt<<endl<<endl;
      (*statFilePtr)<<"op\tbest\tworst\tavg\tbest_#SVars\tworst_#SVars\ta\tb";
      for (i=0;i<GA_OP_NUM;++i) (*statFilePtr)<<"\top"<<i;
      (*statFilePtr)<<endl;
      outputStat();
    }
  
  for (numEpochs=0;numEpochs<maxEpoch && bestScore<goalScore;++numEpochs)
    {
      op=select(accP2);
      switch (op)
	{
	case GA_OP_CROSSOVER:       // crossover

	  s1=select(accP1);
	  s2=select(accP1,s1);

	  if (population[s1].crossover(population[s2],population[worstID1],
				       population[worstID2]))
	    {
	      population[worstID1].evaluate(bestScore,accP2);
	      population[worstID2].evaluate(bestScore,accP2);

	      break;
	    }

	  // if crossover fails mutation is executed!

	case GA_OP_MUTATE:          // mutation

	  s1=select(accP1);
	  if (s1!=worstID1) population[worstID1].assign(population[s1]);

	  population[worstID1].mutate();

	  population[worstID1].evaluate(bestScore,accP2);
	  break;

	case GA_OP_CREEP:           // creep
	  
	  // locally optimize a randomly selected solution
	  s1=select(accP1);
	  sol=population[s1];
	  if (sol.creep())   // is there any improvement?
	    {
	      population[worstID1].assign(sol);
	      population[worstID1].evaluate(bestScore,accP2);	      
	    }
	
	  break;
	}

      computeFitness();
      if ((numEpochs+1)%prtInt==0) outputStat();
    }
}

void Solver::writeSolutions (VFile &solFile)
{
  sort(population.begin(),population.end());
  solFile<<population<<accP2;
}

void Solver::printStat () const
{
  cout<<"* Solver stat report:"<<endl;
  cout<<"  Epochs run = "<<numEpochs<<endl;
  cout<<"  Best score = "<<bestScore<<endl;
  cout<<"  1st worst score = "<<worstScore1<<endl;
  cout<<"  2nd worst score = "<<worstScore2<<endl;
  cout<<"  Avg score = "<<avgScore<<endl<<endl;
}
