// ======================================================================
// solver.hpp - a solver for optimization constraint satisfaction problem
//             using a genetic algorithm
// 
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Added GA operator Creep 
//         (local optimization) - it's problem specific and needed to be
//         'hooked' from outside.
// 110500: Benjamin Han <benhdj@cs.cmu.edu> Support for incremental
//         evaluation based on the changed assignments in a solution is
//         in place
// 110900: Benjamin Han <benhdj@cs.cmu.edu> New Solver::init() added for
//         initializing a population by loading a solution file.
// 111200: Benjamin Han <benhdj@cs.cmu.edu> Change the behavior of
//         crossover - now if no novelty is introduced mutation will take
//         over
// 021601: Benjamin Han <benhdj@cs.cmu.edu> vfile.h -> vFile.hpp.
// 021701: Benjamin Han <benhdj@cs.cmu.edu> Add 'addBest' argument to 
//         Solver::init(); also add Solution::allSatisfied() to set
//         numSVars=Solution::size().
// 080501: Benjamin Han <benhdj@cs.cmu.edu> Minor comments revision.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// ======================================================================

//   Copyright (C) 2000 Benjamin Han <benhdj@cs.cmu.edu>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef SOLVER_HPP
#define SOLVER_HPP

#include <vector>
#include <fstream>

#include <assert.h>

#include "vFile.hpp"
#include "random.hpp"

using namespace Toolbox;
using namespace std;

// ASSUMPTION: possible ranges for vars should not contain UNASSIGNED
#define UNASSIGNED -1

#define GA_OP_CROSSOVER      0
#define GA_OP_MUTATE         1
#define GA_OP_CREEP          2
#define GA_OP_NUM            3         // # of GA operators

struct Var
{
  int vid;   // which variable is this all about?
  Range r;   // the valid range of the values
};

class Solution;

struct Constraints: public vector<Var>
{
  // copy src in random order
  void randomCopy (const Constraints &src);

  // a dumb way to satisfy the constraints and update the solution
  // (read: pure sequential assignments according to the ordering)
  // returns the # of satisfied vars
  int satisfy (Solution &sol);

  // same with satisfy() but no more binary non-equivalence constraints
  int satisfyRelaxed (Solution &sol);
};

// 'SolCreepFunc' is a function object to locally optimize the solution
// The user must derive her own creeping function
// ASSUMPTION: it is NOT necessary to update 'sol.changed'
class SolCreepFunc
{
public:

  // return true if any local improvement is made; otherwise false
  virtual bool operator () (Solution &sol)=0;
};

// 'SolEvalFunc' is a function object to evaluate the solution;
// The user must derive her own evaluation function
// ASSUMPTION: operator () must returns a non-negative value!
class SolEvalFunc
{
public:

  virtual ~SolEvalFunc () {}

  virtual double operator () (const Solution &sol)=0;
};

class Solution: public vector<int>
{
  static int relaxedMode;   // see comments of Solver::init()
  static list<int>::size_type opHistorySize; // how far back in generations should we keep
  static Constraints c0;    // the original constraints for the problem
  static SolCreepFunc *creepFuncPtr;
  static SolEvalFunc *evalFuncPtr;

  #ifndef NDEBUG
  static bool setRelaxedModeF,setOpHistorySizeF,setConstraintsF,
    setCreepFuncPtrF,setEvalFuncPtrF;
  #endif

  int id;                   // a unique ID in the population

  RangeSet assigned;        // assigned values (sorted)
                            // if relaxedMode>0 this is not maintained at all

  int numSVars;             // # of vars satisfied
  double score;

  list<int> opHistory;      // a fixed window of GA operator history for
                            // operator adaptation
                            // the last one is the most recent one

  friend struct Constraints;

  // find all conflicting assignments between both sides of cp and try to
  // satisfy the constraints
  // ASSUMPTION: no conflicts within either side of cp
  void satisfy (size_type cp);

public:

  RangeSet changed;         // set of changed assignments - for incremental
                            // evaluation; NOTE: if sol[i] is changed => i is
                            // in 'changed', but not guaranteed vice versa

  static void setRelaxedMode (int rm)
  { 
    relaxedMode=rm;

    #ifndef NDEBUG
    setRelaxedModeF=true;
    #endif
  }
  static void setOpHistorySize (list<int>::size_type s)
  {
    opHistorySize=s; 

    #ifndef NDEBUG
    setOpHistorySizeF=true;
    #endif
  }
  static void setConstraints (const Constraints &cons) 
  { 
    c0=cons; 

    #ifndef NDEBUG
    setConstraintsF=true;
    #endif
  }
  static void setCreepFunc (SolCreepFunc &cf) 
  { 
    creepFuncPtr=&cf; 

    #ifndef NDEBUG
    setCreepFuncPtrF=true;
    #endif
  }
  static void setEvalFunc (SolEvalFunc &ef) 
  { 
    evalFuncPtr=&ef; 

    #ifndef NDEBUG
    setEvalFuncPtrF=true;
    #endif
  }

  void allSatisfied () { numSVars=size(); }

  static list<int>::size_type getOpHistorySize ()
  { 
    assert(setOpHistorySizeF);
    return opHistorySize; 
  }

  Solution ():numSVars(0) {}

  void setID (int id) { Solution::id=id; }

  // assign() keeps 'id' and updates 'changed' to reflect the differences between
  // 'sol' and itself
  void assign (const Solution &sol);

  // also adapt the operator rates
  double evaluate (double bestScore, PDist &accOpRates);

  // is the i-th solution changed since the last evaluation?
  bool isNotChanged (int i) const { return (changed.find(i)==changed.end()); }

  double getScore () const { return score; }
  int getNumSVars () const { return numSVars; } // how many vars are satisfied?
  int getID () const { return id; }
  
  // one-point crossover:
  // 1. the two parents are NOT changed (except if one of the children happens
  //    to be one of the parents)
  // 2. the two parents are assigned to the children which are
  //    partially swapped, and the conflicting vars are re-satisfied
  //    under non-relaxed mode
  // 3. all the unassigned vars are also satisfied under relaxed mode
  // 4. two children are updated only when the operation yields new
  //    offsprings
  // 5. returns false if no new offsprings can be generated
  bool crossover (const Solution &sol, Solution &child1, Solution &child2) const;

  // one-point guaranteed mutation: if the mutation causes conflict the
  // other conflicting var is reassigned to some other non-conflicting val
  void mutate ();

  // 1. local optimization of the solution: it's problem-specific - the user
  //    must call setCreepFunc() before this operator goes in action
  // 2. returns true is there's any local improvement; otherwise false
  bool creep ();

  void readFromVFile (VFile &vf);
  void writeToVFile (VFile &vf) const;

  // for sorting in descending order
  bool operator < (const Solution &sol) const { return score>sol.score; }
};

ostream &operator << (ostream &s, const Solution &sol);

// Solver implements a genetic algorithm with the following characteristics:
// 1. steady-state with roulette selection;
// 2. only one of the two operators - crossover and mutation -- is invoked in
//    one epoch;
// 3. worst-deletion replacement.
class Solver
{
public:

  typedef vector<Solution>::size_type size_type;

private:

  #ifndef NDEBUG
  bool initFlag;
  #endif

  Solution::size_type solSize;
  vector<Solution> population;
  int bestID,worstID1,worstID2;   // worstID1 is worse than worstID2
  double bestScore,worstScore1,worstScore2,avgScore;
  double a,b;                     // for fitness scaling
  bool printHeader,scaleFail;
  int op, prtInt, numEpochs;

  fstream *statFilePtr;

  // accumulated probability dist: accP1 for population, accP2 for operators
  PDist accP1,accP2;

  // computed fitness values are in accP1 (accumulated)
  // info about average score, best and worst solutions are also updated
  // ASSUMPTION: all solutions are already evaluated
  void computeFitness ();

  // output running stats to file stream '*statFilePtr'
  void outputStat ()
  {
    char opChar;

    switch (op)
      {
      case -1:                   opChar='-'; break;
      case GA_OP_CROSSOVER:      opChar='X'; break;
      case GA_OP_MUTATE:         opChar='M'; break;
      case GA_OP_CREEP:          opChar='C'; break;
      default:                   opChar=0;  // just to suppress the wanring of "use before uninitialized"
      }

    (*statFilePtr)<<opChar<<'\t';
    (*statFilePtr)<<bestScore<<'\t'<<worstScore1<<'\t'<<avgScore<<'\t';
    (*statFilePtr)<<population[bestID].getNumSVars()<<'\t'
		  <<population[worstID1].getNumSVars()<<'\t';
    if (scaleFail) (*statFilePtr)<<1.0<<'\t'<<0.0<<'\t';
    else (*statFilePtr)<<a<<'\t'<<b<<'\t';

    (*statFilePtr)<<accP2[0];
    for (int i=1;i<GA_OP_NUM;++i) (*statFilePtr)<<'\t'<<accP2[i]-accP2[i-1];
    (*statFilePtr)<<endl;
  }

public:

  // init random number generator
  Solver ();

  // 1. ASSUMPTION: all vars in 'c0' have non-empty ranges.
  // 2. if 'relaxedMode' is non-zero the binary non-equivalence constraints
  //    are ignored; mode 2/1 inits the population with/out constraints.
  // 3. 'accOpRates' is an array of accumulated GA operater rates, with the
  //    last element = 0.
  // 4. 'statFile' is used to record the running statistics (every 'prtInt'
  //    epochs).
  // 5. 'opWindowSize' defines how far back in generations the operator credits
  //    are distributed.
  // 6. if a solution is provided 'solFile' then the initial population is
  //    read in from that file instead of being generated randomly
  // 7. if addBest==true the statistically best solution is included in the
  //    init population
  void init (const Constraints &c0, int relaxedMode, list<int>::size_type opWindowSize, 
	     SolCreepFunc &creepFunc, SolEvalFunc &evalFunc, 
	     int prtInt, fstream &statFile, VFile &solFile);
  void init (const Constraints &c0, int relaxedMode, size_type popSize, 
	     const PDist &accOpRates, list<int>::size_type opWindowSize, 
	     SolCreepFunc &creepFunc, SolEvalFunc &evalFunc, 
	     int prtInt, fstream &statFile, bool addBest);

  // run Solver until # of epochs are run *or* the score of the best
  // solution is higher or equal than the goal score
  // ASSUMPTION: The parameters must be positive values!
  void run (int epochs,double goalScore);

  const Solution &getBestSolution () const
    {
      assert(bestID<population.size());
      return population[bestID];
    }

  // write all solutions to solFile in descending order according to the scores;
  void writeSolutions (VFile &solFile);

  void printStat () const;
};


#endif
